
package regression;

import java.util.Enumeration;
import java.util.NoSuchElementException;


/**
 * Interpreted function.
 *
 * @author	Copyright (C) 2002 Alessandro Bollini
 */
public class Function {

	private final String[] tokens; // not null.
	private final Expression expression; // null if unbale to parse tokens.
	private final double range=1e15; // allowed floating point range.


	public Function(final String[] tokens) {

		if ( tokens == null ) {
			throw new IllegalArgumentException("null token array");
		}

		final Enumeration scanner=new Enumeration() {

			private int n=0;

			public boolean hasMoreElements() {
				return n < tokens.length;
			}

			public Object nextElement() {

				try {
					return tokens[n++];
				} catch ( final ArrayIndexOutOfBoundsException exception ) {
					throw new NoSuchElementException("no more elements");
				}
			}

		};

		this.tokens=tokens;
		this.expression=parse(scanner);
	}

	private Expression parse(final Enumeration scanner) {

		try {
			return parseFunction(scanner);
		} catch ( final IllegalArgumentException exception ) {
			return null;
		}
	}

	private Expression parseFunction(final Enumeration scanner) {

		try {

			final Expression expression=parseExpression(scanner);

			if ( scanner.hasMoreElements() ) {
				throw new IllegalArgumentException(
					"unexpected tokens after end of expression");
			}

			return expression;

		} catch ( final NoSuchElementException exception ) {
			throw new IllegalArgumentException(
				"unexpected end of expression");
		}
	}


	private Expression parseExpression(final Enumeration scanner) {

		final String token=(String)scanner.nextElement();

		if ( token.equals("x") ) {

			return new Variable();

		} else if ( token.equals("(") ) {

			final Expression x=parseExpression(scanner);
			final Operator operator=parseOperator(scanner);
			final Expression y=parseExpression(scanner);

			if ( !scanner.nextElement().equals(")") ) {
				throw new IllegalArgumentException(
					"unexpected token <"+token+">");
			}

			return new Operation(operator, x, y);

		} else {
			throw new IllegalArgumentException(
				"unknow token <"+token+">");
		}
	}

	private Operator parseOperator(final Enumeration scanner) {

		final String token=(String)scanner.nextElement();

		if ( token.equals("+") ) {
			return new SumOperator();

		} else if ( token.equals("-") ) {
			return new DifferenceOperator();

		} else if ( token.equals("*") ) {
			return new ProductOperator();

		} else if ( token.equals("/") ) {
			return new RatioOperator();

		} else {
			throw new IllegalArgumentException(
				"unknow operator <"+token+">");
		}
	}


	public double evaluate(final double x) {

		if ( expression == null ) {
			throw new IllegalStateException("undefined expression");
		}

		return expression.evaluate(new Context(x));
	}


	public String toString() {

		final StringBuffer buffer=new StringBuffer();

		for ( int n=0; n < tokens.length; ++n ) {
			if ( n > 0 ) {
				buffer.append(' ');
			}
			buffer.append(tokens[n]);
		}

		return buffer.toString();
	}


	private interface Expression {
		public double evaluate(final Context context);
	}

	private class Context {

		private final double variable;

		public Context(final double x) {
			this.variable=x;
		}

		public double getVariable() {
			return variable;
		}
	}

	private class Variable implements Expression {
		public double evaluate(final Context context) {
			return context.getVariable();
		}
	}

	private class Operation implements Expression {

		private final Operator operator;
		private final Expression x, y;

		public Operation(
			final Operator operator,
			final Expression xExpression,
			final Expression yExpression) {

			this.operator=operator;
			this.x=xExpression;
			this.y=yExpression;
		}

		public double evaluate(final Context context) {
			return operator.process(
				x.evaluate(context),
				y.evaluate(context)
			);
		}
	}


	private interface Operator {
		public double process(final double x, final double y);
	}

	private class SumOperator implements Operator {
		public double process(final double x, final double y) {
			return clip(x+y);
		}
	}

	private class DifferenceOperator implements Operator {
		public double process(final double x, final double y) {
			return clip(x-y);
		}
	}

	private class ProductOperator implements Operator {
		public double process(final double x, final double y) {
			return clip(x*y);
		}
	}

	private class RatioOperator implements Operator {
		public double process(final double x, final double y) {
			return Math.abs(y) > 1.0/range ? clip(x/y)
				: x > 0 ? range : x < 0 ? -range : 0.0;
		}
	}

	private final double clip(final double x) {
		return x > range? range : x < -range ? -range : x;
	}

}
