
package regression;

import java.util.Random;
import java.util.Vector;


/**
 * Evolutionary symbolic regression tool.
 *
 * @author	Copyright (C) 2002 Alessandro Bollini
 */
public class Regressor {

	// Input data.

	private final Data data=new Data();
	private final Grammar grammar=new Grammar();

	// Process configuration.

	private final long seed=System.currentTimeMillis();
	private final Random randomizer=new Random(seed);

	private final int genotypeLengthRange=10;
	private final int phenotypeSizeLimit=100;

	private final int populationSize=100;
	private final int tournamentSize=3;

	private final double recombinationRate=0.7;
	private final double mutationRate=0.1;

	private final int iterationLimit=50;
	private final double tolerance=1e-3;

	// Process state.

	private Regression[] population;

	private int iteration;
	private Regression solution;


	public static void main(final String[] args) {

		if ( args.length > 0 ) {
			throw new IllegalArgumentException("unexpected arguments");
		}

		new Regressor().run();
	}


	private Regressor() {

		this.population=new Regression[populationSize];

		this.iteration=0;
		this.solution=new Regression(
			new byte[0], new Function(new String[0]), Double.MAX_VALUE);

		for ( int n=0; n < population.length; ++n ) {

			final byte[] genotype=generate();
			final Regression individual=develop(genotype);

			this.population[n]=individual;

			if ( individual.getFitness() < solution.getFitness() ) {
				this.solution=individual;
			}
		}
	}

	private byte[] generate() {

		final int length=randomizer.nextInt(genotypeLengthRange-1)+1;
		final byte[] genotype=new byte[length];
		randomizer.nextBytes(genotype);

		return genotype;
	}

	private Regression develop(final byte[] genotype) {

		final Function phenotype=express(genotype);
		final double fitness=evaluate(phenotype);

		return new Regression(genotype, phenotype, fitness);
	}

	private Function express(final byte[] genotype) {

		final String[] tokens=translate(genotype);
		final Function phenotype=new Function(tokens);

		return phenotype;
	}

	private String[] translate(final byte[] genotype) {

		final Grammar.Symbol root=grammar.getRoot();

		final Vector derivation=new Vector();
		derivation.add(root);

		for ( int n=0, locus=0; n < derivation.size(); ) {

			final Object token=derivation.get(n);

			if ( token instanceof Grammar.Symbol ) {

				final Grammar.Symbol symbol=(Grammar.Symbol)token;
				final int options=symbol.getOptions();

				if ( options > 0 && locus < genotype.length ) {

					final int option=(genotype[locus++]&0xFF)%options;
					final Object[] expansion=symbol.getRule(option);

					derivation.remove(n);

					for ( int k=0; k < expansion.length; ++k ) {
						derivation.add(n+k, expansion[k]);
					}

				} else {
					return new String[0];
				}

			} else {
				++n;
			}

			if ( derivation.size() > phenotypeSizeLimit ) {
				return new String[0];
			}
		}

		final String[] tokens=new String[derivation.size()];
		derivation.copyInto(tokens);

		return tokens;
	}

	private double evaluate(final Function phenotype) {

		try {

			final int points=data.getPointCount();

			double error=0.0;

			for ( int n=0; n < points; ++n ) {

				final double x=data.getX(n);
				final double y=data.getY(n);
				final double delta=phenotype.evaluate(x)-y;

				error+=delta*delta;
			}

			return !Double.isNaN(error) && !Double.isInfinite(error) ?
				Math.sqrt(error/points) : Double.MAX_VALUE;

		} catch ( final IllegalStateException exception ) {
			return Double.MAX_VALUE;

		} catch ( final RuntimeException exception ) {
			return Double.MAX_VALUE;
		}
	}


	private void run() {

		System.out.println("------------------------------ Configuration --");
		System.out.println("seed\t"+seed);

		System.out.println("-------------------------------------- Trace --");
		System.out.println("iter.\terror\texpression");

		while ( !completed() ) {

			System.out.println(
				iteration
				+"\t\t"+round(solution.getFitness())
				+"\t"+solution.getPhenotype()
			);

			iterate();
		}

		System.out.println("--------------------------------- Regression --");
		System.out.println("error\t\t"+round(solution.getFitness()));
		System.out.println("expression\t"+solution.getPhenotype());
	}

	private double round(final double value) {
		return ((int)(1000.0*value))/1000.0;
	}


	private boolean completed() {
		return solution.getFitness() < tolerance
			|| iteration >= iterationLimit;
	}

	private void iterate() {

		for ( int n=0; n < population.length; ++n ) {

			final Regression parent=select();
			final Regression offspring=reproduce(parent);

			final int target=randomizer.nextInt(population.length);
			population[target]=offspring;

			if ( offspring.getFitness() < solution.getFitness() ) {
				solution=offspring;
			}
		}

		++iteration;
	}


	private Regression select() {

		Regression winner=sample();

		for ( int n=0; n < tournamentSize; ++n ) {

			final Regression challenger=sample();

			if ( challenger.getFitness() < winner.getFitness() ) {
				winner=challenger;
			}
		}

		return winner;
	}

	private Regression sample() {
		return population[randomizer.nextInt(population.length)];
	}


	private Regression reproduce(final Regression individual) {

		final double event=randomizer.nextDouble();
		final byte[] genotype;

		if ( event < recombinationRate ) {

			final Regression peer=select();

			genotype=recombine(
				individual.getGenotype(), peer.getGenotype());

		} else if ( event-recombinationRate < mutationRate ) {

			genotype=mutate(individual.getGenotype());

		} else {

			genotype=replicate(individual.getGenotype());
		}

		return develop(genotype);
	}

	private byte[] recombine(
		final byte[] targetGenotype, byte[] sourceGenotype) {

		final int headLength=(targetGenotype.length > 1)?
			randomizer.nextInt(targetGenotype.length-1)+1 : 1;
		final int tailLenght=(sourceGenotype.length > 1)?
			randomizer.nextInt(sourceGenotype.length-1)+1 : 1;
		final byte[] combinedGenotype=new byte[headLength+tailLenght];

		System.arraycopy(targetGenotype, 0,
			combinedGenotype, 0, headLength);

		System.arraycopy(sourceGenotype, sourceGenotype.length-tailLenght,
			combinedGenotype, headLength, tailLenght);

		return combinedGenotype;

	}

	private byte[] mutate(final byte[] targetGenotype) {

		final int locus=randomizer.nextInt(targetGenotype.length);
		final byte[] mutatedGenotype=(byte[])targetGenotype.clone();

		mutatedGenotype[locus]=(byte)(randomizer.nextInt()&0xFF);

		return mutatedGenotype;
	}

	private byte[] replicate(final byte[] targetGenotype) {
		return (byte[])targetGenotype.clone();
	}
}
